# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

from types import SimpleNamespace

import pytest
import requests

from galaxy_importer import exceptions as exc
from galaxy_importer.ansible_test.runners import openshift_job


@pytest.fixture
def job():
    return openshift_job.Job(
        'my_domain', 'my_ns', 'session_token', 'ca_path', 'image', 'job_template', logger=None)


@pytest.fixture
def build():
    return openshift_job.Build(
        'my_domain', 'my_ns', 'session_token', 'ca_path',
        'build_template', 'archive_url', logger=None)


def test_runner_run(mocker, monkeypatch):
    mocker.patch.object(openshift_job.OpenshiftJobTestRunner, '_get_token')
    mocker.patch.object(openshift_job.OpenshiftJobTestRunner, '_get_job_template')
    mocker.patch.object(openshift_job.OpenshiftJobTestRunner, '_get_pulp_archive_url')
    mocker.patch.object(openshift_job.Job, 'create')
    mocker.patch.object(openshift_job.Job, 'wait_on_pod_ready')
    mocker.patch.object(openshift_job.Job, 'get_logs')
    mocker.patch.object(openshift_job.Job, 'cleanup')
    mocker.patch.object(openshift_job.Build, 'start_and_get_image_link')
    mocker.patch.object(openshift_job.Build, 'cleanup')

    openshift_job.Job.get_logs.return_value = ['log_entry', b'bytes_log_entry']
    openshift_job.OpenshiftJobTestRunner._get_pulp_archive_url.return_value = 'image_link'
    monkeypatch.setenv('IMPORTER_API_DOMAIN', 'my_host')
    monkeypatch.setenv('IMPORTER_JOB_NAMESPACE', 'my_project')
    runner = openshift_job.OpenshiftJobTestRunner()
    runner.run()

    assert openshift_job.Build.start_and_get_image_link.called
    assert openshift_job.Job.create.called
    assert openshift_job.Job.wait_on_pod_ready.called
    assert openshift_job.Job.get_logs.called
    assert openshift_job.Job.cleanup.called
    assert openshift_job.Build.cleanup.called


def test_runner_get_token(mocker, tmp_path):
    mocker.patch.object(openshift_job, 'OCP_SERVICEACCOUNT_PATH')
    p = tmp_path / 'token'
    p.write_text('my_session_token_1234')
    openshift_job.OCP_SERVICEACCOUNT_PATH = str(tmp_path) + '/'
    assert openshift_job.OpenshiftJobTestRunner._get_token() == 'my_session_token_1234'


def test_runner_get_job_template(mocker, tmp_path):
    job_template = openshift_job.OpenshiftJobTestRunner._get_job_template()
    assert job_template.startswith('apiVersion: batch/v1\nkind: Job')


def test_job_init(job):
    assert job.jobs_url == 'my_domain/apis/batch/v1/namespaces/my_ns/jobs'
    assert job.pods_url == 'my_domain/api/v1/namespaces/my_ns/pods'


def test_job_create(mocker, job):
    mocker.patch.object(requests, 'post')

    requests.post.return_value = SimpleNamespace(status_code=201)
    job.create()
    assert requests.post.called

    requests.post.return_value = SimpleNamespace(status_code=500, reason='', text='')
    with pytest.raises(exc.AnsibleTestError):
        job.create()


def test_job_wait_on_pod_ready(mocker, job):
    mocker.patch.object(openshift_job.Job, 'get_pods')
    mocker.patch.object(openshift_job.Job, 'cleanup')
    mocker.patch.object(openshift_job, 'API_CHECK_RETRIES')
    openshift_job.API_CHECK_RETRIES = 1

    openshift_job.Job.get_pods.return_value = [{'status': {'phase': 'Running'}}]
    job.wait_on_pod_ready()
    assert openshift_job.Job.get_pods.called

    openshift_job.Job.get_pods.return_value = []
    with pytest.raises(exc.AnsibleTestError):
        job.wait_on_pod_ready()
    assert openshift_job.Job.cleanup.called

    openshift_job.Job.get_pods.return_value = [{'status': {'phase': 'Pending'}}]
    with pytest.raises(exc.AnsibleTestError):
        job.wait_on_pod_ready()
    assert openshift_job.Job.cleanup.called


def test_job_get_pods(mocker, job):
    mocker.patch.object(requests, 'get')
    requests.get.return_value.json.return_value = {'items': []}
    job.get_pods()
    assert requests.get.called


def test_job_get_pods_fail(mocker, job):
    mocker.patch.object(requests, 'get')
    requests.get.return_value.json.return_value = {}
    with pytest.raises(exc.AnsibleTestError, match=r'Could not access pod assocated with job'):
        job.get_pods()
    assert requests.get.called


def test_job_get_pod_name():
    pod = {'metadata': {'name': 'my_pod_name'}}
    assert openshift_job.Job.get_pod_name(pod) == 'my_pod_name'


def test_job_get_logs(mocker, job):
    mocker.patch.object(openshift_job.Job, 'get_pods')
    mocker.patch.object(requests, 'get')
    openshift_job.Job.get_pods.return_value = [{'metadata': {'name': 'my_pod_name'}}]
    requests.get.iter_lines.return_value = {}
    job.get_logs()
    assert requests.get.called


def test_job_cleanup(mocker, job):
    mocker.patch.object(openshift_job.Job, 'get_pods')
    mocker.patch.object(requests, 'delete')
    openshift_job.Job.get_pods.return_value = [{
        'metadata': {'name': 'my_pod_name'},
        'status': {'phase': 'Succeeded'},
    }]
    job.cleanup()
    assert requests.delete.called


def test_job_cleanup_fail(mocker, job):
    mocker.patch.object(openshift_job.Job, 'get_pods')
    mocker.patch.object(requests, 'delete')
    openshift_job.Job.get_pods.return_value = [{
        'metadata': {'name': 'my_pod_name'},
        'status': {
            'phase': 'Failed',
            'containerStatuses': [
                {'state': {'terminated': {'reason': 'internal_error_123'}}}
            ]
        },
    }]
    with pytest.raises(
        exc.AnsibleTestError,
        match=r'Pod terminated with status: "Failed" and reason: "internal_error_123"'
    ):
        job.cleanup()
    assert requests.delete.called


def test_build_create_buildconfig(mocker, build):
    mocker.patch.object(requests, 'post')

    requests.post.return_value = SimpleNamespace(status_code=201)
    build._create_buildconfig()
    assert requests.post.called

    requests.post.return_value = SimpleNamespace(status_code=500, reason='', text='')
    with pytest.raises(exc.AnsibleTestError):
        build._create_buildconfig()


def test_job_get_build(mocker, build):
    mocker.patch.object(requests, 'get')

    requests.get.return_value = SimpleNamespace(status_code=200, json=lambda: {'func': 'result'})
    assert build._get_build() == {'func': 'result'}

    requests.get.return_value = SimpleNamespace(status_code=500, reason='', text='')
    with pytest.raises(exc.AnsibleTestError):
        build._get_build()


def test_job_get_image(mocker, build):
    mocker.patch.object(requests, 'get')

    requests.get.return_value = SimpleNamespace(status_code=200, json=lambda: {'func': 'result'})
    assert build._get_image() == {'func': 'result'}

    requests.get.return_value = SimpleNamespace(status_code=500)
    assert build._get_image() is None


def test_runner_get_build_template(mocker, tmp_path):
    build_template = openshift_job.OpenshiftJobTestRunner._get_build_template()
    assert build_template.startswith('apiVersion: build.openshift.io/v1\nkind: BuildConfig')


def test_wait_until_build_created(mocker, build):
    mocker.patch.object(openshift_job.Build, '_get_build')
    mocker.patch.object(openshift_job, 'API_CHECK_RETRIES')
    openshift_job.API_CHECK_RETRIES = 1

    openshift_job.Build._get_build.return_value = {'items': ['item']}
    build._wait_until_build_created()

    openshift_job.Build._get_build.return_value = {'items': []}
    with pytest.raises(exc.AnsibleTestError):
        build._wait_until_build_created()


def test_wait_until_build_complete(mocker, build):
    mocker.patch.object(openshift_job.Build, '_get_build')
    mocker.patch.object(openshift_job, 'API_CHECK_RETRIES')
    openshift_job.API_CHECK_RETRIES = 1

    openshift_job.Build._get_build.return_value = {'items': [{'status': {'phase': 'Complete'}}]}
    build._wait_until_build_complete()

    openshift_job.Build._get_build.return_value = {'items': [{'status': {'phase': 'Pending'}}]}
    with pytest.raises(exc.AnsibleTestError):
        build._wait_until_build_complete()


def test_wait_until_image_available(mocker, build):
    mocker.patch.object(openshift_job.Build, '_get_image')
    mocker.patch.object(openshift_job, 'API_CHECK_RETRIES')
    openshift_job.API_CHECK_RETRIES = 1

    openshift_job.Build._get_image.return_value = 'image_link'
    build._wait_until_image_available()

    openshift_job.Build._get_image.return_value = None
    with pytest.raises(exc.AnsibleTestError):
        build._wait_until_image_available()


def test_delete_buildconfig(mocker, build):
    mocker.patch.object(requests, 'delete')
    requests.delete.return_value = SimpleNamespace(status_code=204, reason='', text='')
    build._delete_buildconfig()
    assert requests.delete.called


def test_delete_imagestreamtag(mocker, build):
    mocker.patch.object(requests, 'delete')
    requests.delete.return_value = SimpleNamespace(status_code=204, reason='', text='')
    build._delete_imagestreamtag()
    assert requests.delete.called


def test_build_cleanup(mocker, build):
    mocker.patch.object(openshift_job.Build, '_delete_buildconfig')
    mocker.patch.object(openshift_job.Build, '_delete_imagestreamtag')
    mocker.patch.object(requests, 'delete')
    build.cleanup()
    assert openshift_job.Build._delete_buildconfig.called
    assert openshift_job.Build._delete_imagestreamtag.called


def test_build_start_and_get_image_link(mocker, build):
    mocker.patch.object(openshift_job.Build, '_create_buildconfig')
    mocker.patch.object(openshift_job.Build, '_wait_until_build_created')
    mocker.patch.object(openshift_job.Build, '_wait_until_build_complete')
    mocker.patch.object(openshift_job.Build, '_wait_until_image_available')
    mocker.patch.object(openshift_job.Build, '_get_image')
    openshift_job.Build._get_image.return_value = {
        'metadata': {'name': 'image_stream_tag_name'}}
    assert build.start_and_get_image_link() == 'image_stream_tag_name'
    assert openshift_job.Build._create_buildconfig.called
    assert openshift_job.Build._wait_until_build_created.called
    assert openshift_job.Build._wait_until_build_complete.called
    assert openshift_job.Build._wait_until_image_available.called



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import os

import pytest

from galaxy_importer import config


@pytest.fixture
def temp_config_file():
    try:
        dir = os.path.dirname(os.path.dirname(__file__))
        config_file = os.path.join(dir, 'galaxy_importer', 'galaxy-importer.cfg')
        yield config_file
    finally:
        os.remove(config_file)


def test_config_set_from_file(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True\n'
                'INFRA_PULP = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert cfg.log_level_main == 'INFO'
        assert cfg.run_ansible_test is True
        assert cfg.infra_pulp is True
        assert cfg.infra_osd is False


def test_config_no_file():
    config_data = config.ConfigFile.load()
    assert not config_data


def test_no_config_data():
    cfg = config.Config(config_data={})
    assert cfg.log_level_main == 'INFO'
    assert cfg.run_ansible_test is False
    assert cfg.infra_pulp is False
    assert cfg.infra_osd is False


def test_config_bad_ini_section(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[bad-section]\nRUN_ANSIBLE_TEST = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert cfg.log_level_main == 'INFO'
        assert cfg.run_ansible_test is False
        assert cfg.infra_pulp is False
        assert cfg.infra_osd is False


def test_config_with_non_boolean(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True\n'
                'LOG_LEVEL_MAIN = DEBUG')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert cfg.log_level_main == 'DEBUG'
        assert cfg.run_ansible_test is True
        assert cfg.infra_pulp is False
        assert cfg.infra_osd is False



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.


import logging
import os
import pkg_resources
import requests
import time
import uuid
import yaml

from galaxy_importer import config
from galaxy_importer import exceptions
from galaxy_importer.ansible_test.runners.base import BaseTestRunner


default_logger = logging.getLogger(__name__)

cfg = config.Config()
API_CHECK_RETRIES = 300
API_CHECK_DELAY_SECONDS = 1
OCP_SERVICEACCOUNT_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/'
IMAGE_BASE_NAME = 'ansible-test'


class OpenshiftJobTestRunner(BaseTestRunner):
    """Run image as an openshift job."""
    def run(self):
        self.log.info('Preparing build and job pod to run ansible-test sanity')
        openshift_build = Build(
            ocp_domain=os.environ['IMPORTER_API_DOMAIN'],
            namespace=os.environ['IMPORTER_JOB_NAMESPACE'],
            session_token=OpenshiftJobTestRunner._get_token(),
            ca_path=OpenshiftJobTestRunner._get_ca_path(),
            archive_url=self._get_pulp_archive_url(self.file),
            build_template=OpenshiftJobTestRunner._get_build_template(),
            logger=self.log,
        )

        image_link = openshift_build.start_and_get_image_link()
        self.log.debug(f'image_link={image_link}')

        job = Job(
            ocp_domain=os.environ['IMPORTER_API_DOMAIN'],
            namespace=os.environ['IMPORTER_JOB_NAMESPACE'],
            session_token=OpenshiftJobTestRunner._get_token(),
            ca_path=OpenshiftJobTestRunner._get_ca_path(),
            image=image_link,
            job_template=OpenshiftJobTestRunner._get_job_template(),
            logger=self.log,
        )
        job.create()
        job.wait_on_pod_ready()
        iter_logs = job.get_logs()

        for line in iter_logs:
            if isinstance(line, bytes):
                self.log.error('Unexpected bytes in logs: {}'.format(str(line)))
                continue
            self.log.info(line)

        openshift_build.cleanup()
        job.cleanup()

    @staticmethod
    def _get_token():
        with open(os.path.join(OCP_SERVICEACCOUNT_PATH, 'token'), 'r') as f:
            token = f.read().rstrip()
        return token

    @staticmethod
    def _get_ca_path():
        return os.path.join(OCP_SERVICEACCOUNT_PATH, 'ca.crt')

    @staticmethod
    def _get_job_template():
        path = pkg_resources.resource_filename('galaxy_importer', 'ansible_test/job_template.yaml')
        with open(path, 'r') as f:
            job_template = f.read()
        return job_template

    @staticmethod
    def _get_build_template():
        path = pkg_resources.resource_filename(
            'galaxy_importer', 'ansible_test/build_template.yaml')
        with open(path, 'r') as f:
            build_template = f.read()
        return build_template

    @staticmethod
    def _get_pulp_archive_url(pulp_artifact_file):
        parameters = {'ResponseContentDisposition': 'attachment;filename=archive.tar.gz'}
        return pulp_artifact_file.storage.url(pulp_artifact_file.name, parameters=parameters)


class Build(object):
    """Interact with Openshift builds via REST API."""
    def __init__(self, ocp_domain, namespace, session_token,
                 ca_path, build_template, archive_url, logger):
        self.name = 'import-' + str(uuid.uuid4())
        self.auth_header = {'Authorization': f'Bearer {session_token}'}
        self.ca_path = ca_path
        self.api_build_prefix = f'{ocp_domain}/apis/build.openshift.io/v1/namespaces/{namespace}'
        self.buildconfig_url = f'{self.api_build_prefix}/buildconfigs'
        self.build_url = f'{self.api_build_prefix}/builds'
        self.istag_url = \
            f'{ocp_domain}/apis/image.openshift.io/v1/namespaces/{namespace}/imagestreamtags'
        self.image_name = f'{IMAGE_BASE_NAME}:{self.name}'
        self.build_yaml = build_template.format(
            image_name=self.image_name,
            build_name=self.name,
            download_url=archive_url,
        )
        self.log = logger or default_logger

    def start_and_get_image_link(self):
        self._create_buildconfig()

        self._wait_until_build_created()
        self._wait_until_build_complete()
        self._wait_until_image_available()

        imagestream_tag = self._get_image()
        return imagestream_tag['metadata']['name']

    def _create_buildconfig(self):
        self.log.info(f'Creating buildconfig {self.name}')
        r = requests.post(
            self.buildconfig_url,
            headers=self.auth_header,
            json=yaml.safe_load(self.build_yaml),
            verify=self.ca_path,
        )
        if r.status_code != requests.codes.created:
            raise exceptions.AnsibleTestError(
                f'Could not create buildconfig: {r.status_code} {r.reason} {r.text}')

    def _get_build(self):
        """Get build associated with buildconfig."""
        params = {'labelSelector': f'buildconfig={self.name}'}
        r = requests.get(
            self.build_url, headers=self.auth_header, params=params, verify=self.ca_path)
        if r.status_code != requests.codes.ok:
            raise exceptions.AnsibleTestError('Could not access builds')
        return r.json()

    def _get_image(self):
        """Get imagestream tag associated with buildconfig."""
        r = requests.get(
            '{}/{}'.format(self.istag_url, self.image_name),
            headers=self.auth_header,
            verify=self.ca_path)
        if r.status_code != requests.codes.ok:
            return None
        return r.json()

    def _wait_until_build_created(self):
        self.log.info('Creating build...')
        for i in range(API_CHECK_RETRIES):
            build = self._get_build()
            if len(build['items']) > 0:
                return
            time.sleep(API_CHECK_DELAY_SECONDS)

        raise exceptions.AnsibleTestError('Could not create build')

    def _wait_until_build_complete(self):
        self.log.info('Waiting until build is complete...')
        for i in range(API_CHECK_RETRIES):
            build = self._get_build()
            build_phase = build['items'][0]['status']['phase']
            if build_phase == 'Complete':
                return
            time.sleep(API_CHECK_DELAY_SECONDS)

        raise exceptions.AnsibleTestError('Unable to build image within timeout')

    def _wait_until_image_available(self):
        self.log.info('Waiting until image is available...')
        for i in range(API_CHECK_RETRIES):
            if self._get_image():
                return
            time.sleep(API_CHECK_DELAY_SECONDS)
        raise exceptions.AnsibleTestError('Image was not available')

    def _delete_buildconfig(self):
        r = requests.delete(
            '{}/{}'.format(self.buildconfig_url, self.name),
            headers=self.auth_header,
            verify=self.ca_path)
        self.log.debug(f'delete bc: {r.status_code} {r.reason} {r.text}')

    def _delete_imagestreamtag(self):
        r = requests.delete(
            '{}/{}'.format(self.istag_url, self.image_name),
            headers=self.auth_header,
            verify=self.ca_path)
        self.log.debug(f'delete istag: {r.status_code} {r.reason} {r.text}')

    def cleanup(self):
        self._delete_buildconfig()
        self._delete_imagestreamtag()


class Job(object):
    """Interact with Openshift Job via REST API."""

    def __init__(self, ocp_domain, namespace, session_token, ca_path, image, job_template, logger):
        self.name = 'ansible-test-' + str(uuid.uuid4())
        self.auth_header = {'Authorization': f'Bearer {session_token}'}
        self.ca_path = ca_path
        self.jobs_url = f'{ocp_domain}/apis/batch/v1/namespaces/{namespace}/jobs'
        self.job_name_url = f'{self.jobs_url}/{self.name}'
        self.pods_url = f'{ocp_domain}/api/v1/namespaces/{namespace}/pods'
        self.pod_url_template = '{pods_url}/{pod_name}'
        self.log_url_template = '{pods_url}/{pod_name}/log'
        self.job_yaml = job_template.format(
            job_name=self.name,
            image=image,
            memory_request=os.environ.get('IMPORTER_MEMORY_REQUEST', '256Mi'),
            memory_limit=os.environ.get('IMPORTER_MEMORY_LIMIT', '1Gi'),
            cpu_request=os.environ.get('IMPORTER_CPU_REQUEST', '500m'),
            cpu_limit=os.environ.get('IMPORTER_CPU_LIMIT', '500m'),
        )
        self.log = logger or default_logger

    def create(self):
        """Create the job."""

        self.log.info(f'Creating job {self.name}')
        r = requests.post(
            self.jobs_url,
            headers=self.auth_header,
            json=yaml.safe_load(self.job_yaml),
            verify=self.ca_path,
        )
        if r.status_code != requests.codes.created:
            raise exceptions.AnsibleTestError(
                f'Could not create job: {r.status_code} {r.reason} {r.text}')

    def wait_on_pod_ready(self):
        """Wait until job's pod initializes, pulls image, and starts running."""

        self.log.info('Creating pod...')
        for i in range(API_CHECK_RETRIES):
            pods = self.get_pods()
            if len(pods) < 1:
                time.sleep(API_CHECK_DELAY_SECONDS)
                continue
            break

        if len(pods) < 1:
            self.cleanup()
            raise exceptions.AnsibleTestError('Could not create pod associated with job')

        self.log.info('Scheduling pod and waiting until it is running...')
        for i in range(API_CHECK_RETRIES):
            pods = self.get_pods()
            pod_phase = pods[0]['status']['phase']
            if pod_phase != 'Pending':
                return
            time.sleep(API_CHECK_DELAY_SECONDS)

        self.log.debug(pods[0]['status'])
        self.cleanup()
        raise exceptions.AnsibleTestError('Could not start pod associated with job')

    def get_pods(self):
        """Get pods associated with job."""
        params = {'labelSelector': f'job-name={self.name}'}
        r = requests.get(
            self.pods_url, headers=self.auth_header, params=params, verify=self.ca_path)
        try:
            pods = r.json()['items']
        except (KeyError, ValueError):
            raise exceptions.AnsibleTestError('Could not access pod assocated with job')
        return pods

    @staticmethod
    def get_pod_name(pod):
        return pod['metadata']['name']

    def get_logs(self):
        """Returns stream of lines from the logs of the pod."""
        pod = self.get_pods()[0]
        pod_name = self.get_pod_name(pod)
        r = requests.get(
                url=f'{self.pods_url}/{pod_name}/log',
                headers=self.auth_header,
                params=dict(follow='true'),
                verify=self.ca_path,
                stream=True,
            )
        return r.iter_lines(decode_unicode=True)

    def cleanup(self):
        """Deletes job and any pods associated to it."""
        requests.delete(self.job_name_url, headers=self.auth_header, verify=self.ca_path)
        self.log.debug(f'Deleted job {self.name}')

        for pod in self.get_pods():
            pod_name = self.get_pod_name(pod)
            requests.delete(
                f'{self.pods_url}/{pod_name}', headers=self.auth_header, verify=self.ca_path)
            self.log.debug(f'Deleted pod {pod_name}')

            status = pod['status']['phase']
            if status == 'Failed':
                reason = pod['status']['containerStatuses'][0]['state']['terminated']['reason']
                raise exceptions.AnsibleTestError(
                    f'Pod terminated with status: "{status}" and reason: "{reason}"')



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.


import os
import subprocess

from galaxy_importer.ansible_test.runners.base import BaseTestRunner


class LocalAnsibleTestRunner(BaseTestRunner):
    """Run ansible-test locally with --docker or using venv."""
    def run(self):
        version_proc = subprocess.Popen(
            ['ansible', '--version'],
            stdout=subprocess.PIPE,
            encoding='utf-8',
        )
        self.log.info(f'Using {list(version_proc.stdout)[0].rstrip()}')

        suffix = f'ansible_collections/{self.metadata.namespace}/{self.metadata.name}/'
        collection_dir = os.path.join(self.dir, suffix)

        cmd = [
            'ansible-test', 'sanity',
            '--docker',
            '--color', 'yes',
            '--failure-ok',
        ]

        collection_name = f'{self.metadata.namespace}-{self.metadata.name}-{self.metadata.version}'
        self.log.info(f'Running ansible-test sanity on {collection_name} ...')
        self.log.info(f'{" ".join(cmd)}')

        proc = subprocess.Popen(
            cmd,
            cwd=collection_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding='utf-8',
        )

        for line in proc.stdout:
            self.log.info(line.strip())

        return_code = proc.wait()
        if return_code != 0:
            self.log.error('An exception occurred in {}, returncode={}, collection={}'.format(
                ' '.join(cmd), return_code, collection_name))



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import argparse
import json
import logging
import os
import re
import sys

from galaxy_importer import collection
from galaxy_importer import config
from galaxy_importer.exceptions import ImporterError

FILENAME_REGEXP = re.compile(
    r"^(?P<namespace>\w+)-(?P<name>\w+)-"
    r"(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$"
)


def main(args=None):
    config_data = config.ConfigFile.load()
    cfg = config.Config(config_data=config_data)
    setup_logger(cfg)
    args = parse_args(args)

    data = call_importer(filepath=args.file, cfg=cfg)
    if not data:
        return

    if args.print_result:
        print(json.dumps(data, indent=4))

    write_output_file(data)


def setup_logger(cfg):
    """Sets up logger with custom formatter."""
    logger = logging.getLogger()
    logger.setLevel(getattr(logging, cfg.log_level_main, 'INFO'))

    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)


class CustomFormatter(logging.Formatter):
    """Formatter that does not display INFO loglevel."""
    def formatMessage(self, record):
        if record.levelno == logging.INFO:
            return '%(message)s' % vars(record)
        else:
            return '%(levelname)s: %(message)s' % vars(record)


def parse_args(args):
    parser = argparse.ArgumentParser(
        description='Run importer on collection and save result to disk.')
    parser.add_argument(
        'file',
        help='artifact to import')
    parser.add_argument(
        '--print-result',
        dest='print_result',
        action='store_true',
        help='print importer result to console')
    return parser.parse_args(args=args)


def call_importer(filepath, cfg):
    """Returns result of galaxy_importer import process.

    :param file: Artifact file to import.
    """
    match = FILENAME_REGEXP.match(os.path.basename(filepath))
    namespace, name, version = match.groups()
    filename = collection.CollectionFilename(namespace, name, version)

    with open(filepath, 'rb') as f:
        try:
            data = collection.import_collection(f, filename, cfg=cfg)
        except ImporterError as e:
            logging.error(f'The import failed for the following reason: {str(e)}')
            return None
        except Exception:
            logging.error('Unexpected error occurred:', exc_info=True)
            return None

    logging.info('Importer processing completed successfully')
    return data


def write_output_file(data):
    with open('importer_result.json', 'w') as output_file:
        output_file.write(json.dumps(data, indent=4))


if __name__ == '__main__':
    exit(main())



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import os
import shutil
import tempfile
import unittest

from galaxy_importer.finder import ContentFinder


class TestContentFinder(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

        os.mkdir(os.path.join(self.temp_dir, 'plugins'))
        self.module_dir = os.path.join(self.temp_dir, 'plugins', 'modules')
        os.mkdir(self.module_dir)

        self.roles_dir = os.path.join(self.temp_dir, 'roles')
        os.mkdir(self.roles_dir)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_find_content(self):
        with open(os.path.join(self.module_dir, '__init__.py'), 'w'):
            pass
        with open(os.path.join(self.module_dir, 'first_module.py'), 'w'):
            pass
        with open(os.path.join(self.module_dir, 'second_module.py'), 'w'):
            pass

        my_roles_dir = os.path.join(self.roles_dir, 'my_role')
        os.mkdir(my_roles_dir)
        os.mkdir(os.path.join(my_roles_dir, 'tasks'))

        contents = ContentFinder().find_contents(self.temp_dir)

        content_items = [os.path.basename(c.path) for c in contents]
        assert 'first_module.py' in content_items
        assert 'second_module.py' in content_items
        assert '__init__.py' not in content_items
        assert 'my_role' in content_items
        assert len(content_items) == 3

    def test_find_no_content(self):
        contents = ContentFinder().find_contents(self.temp_dir)
        assert not any(True for _ in contents)

    def test_skip_plugin_files(self):
        with open(os.path.join(self.module_dir, '__init__.py'), 'w'):
            pass
        with open(os.path.join(self.module_dir, 'main.go'), 'w'):
            pass
        contents = ContentFinder().find_contents(self.temp_dir)
        assert not any(True for _ in contents)

    def test_nested_plugin(self):
        subdir1 = os.path.join(self.module_dir, 'subdir1')
        os.mkdir(subdir1)
        subdir2 = os.path.join(subdir1, 'subdir2')
        os.mkdir(subdir2)
        with open(os.path.join(subdir2, 'nested_module.py'), 'w'):
            pass

        contents = ContentFinder().find_contents(self.temp_dir)
        assert list(contents)[0].path == \
            'plugins/modules/subdir1/subdir2/nested_module.py'

    def test_nested_role(self):
        subdir1 = os.path.join(self.roles_dir, 'subdir1')
        os.mkdir(subdir1)
        role_dir = os.path.join(subdir1, 'my_role')
        os.mkdir(role_dir)

        contents = ContentFinder().find_contents(self.temp_dir)
        assert len(contents) == 0

        dir_in_role = os.path.join(role_dir, 'tasks')
        os.mkdir(dir_in_role)

        contents = ContentFinder().find_contents(self.temp_dir)
        assert list(contents)[0].path == \
            'roles/subdir1/my_role'

    def test_error_file_in_roles_dir(self):
        with open(os.path.join(self.roles_dir, 'main.yml'), 'w'):
            pass
        my_roles_dir = os.path.join(self.roles_dir, 'my_role')
        os.mkdir(my_roles_dir)
        os.mkdir(os.path.join(my_roles_dir, 'tasks'))

        contents = ContentFinder().find_contents(self.temp_dir)
        content_items = [os.path.basename(c.path) for c in contents]
        assert 'my_role' in content_items
        assert len(content_items) == 1



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import collections
import os
from unittest import mock

from pyfakefs.fake_filesystem_unittest import TestCase
import pytest

from galaxy_importer.utils import markup as markup_utils

TEXT_SIMPLE = 'A simple description'
TEXT_BAD_TAG = 'hello <script>cruel</script> world'
TEXT_BAD_HTML_IN_MD = '''
> hello <a name="n"
> href="javascript:something_bad()">*you*</a>
'''
TEXT_FORMATTING = '''
# Role

[Tool](https://www.example.com) installed in `$PATH`

### Installation

    package_version: "1.2.0"

> NOTE: Tool 'feature' is _beta_

* Item1
* Item2
'''

DocFile = collections.namedtuple('DocFile', ['text', 'mimetype'])


class TestFindGetFiles(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.directory = '/tmp'

    def test_find_readme(self):
        assert not markup_utils._find_readme(self.directory)
        file_path = os.path.join(self.directory, 'README.md')
        self.fs.create_file(file_path)
        res = markup_utils._find_readme(self.directory)
        assert res == '/tmp/README.md'

    def test_cannot_find_invalid_readmes(self):
        for invalid_name in [
            'readme.md',
            'README',
            'README.rst',
            'readme',
            'INTRO.md',
        ]:
            self.fs.create_file(os.path.join(self.directory, invalid_name))

        assert not markup_utils._find_readme(self.directory)

    def test_get_readme_doc_file(self):
        res = markup_utils.get_readme_doc_file(self.directory)
        assert res is None

        file_path = os.path.join(self.directory, 'README.md')
        self.fs.create_file(file_path, contents='## My Collection')
        res = markup_utils.get_readme_doc_file(self.directory)
        assert res.name == 'README.md'
        assert res.text == '## My Collection'

    def test_get_file(self):
        filename = 'README.md'
        res = markup_utils._get_file(self.directory, filename)
        assert res is None

        file_path = os.path.join(self.directory, filename)
        self.fs.create_file(file_path)
        res = markup_utils._get_file(self.directory, file_path)
        assert res.name == 'README.md'

    @mock.patch('os.path.getsize')
    def test_get_file_with_file_size_error(self, getsize):
        getsize.return_value = (512 ** 2) + 1
        filename = 'README.md'
        file_path = os.path.join(self.directory, filename)
        self.fs.create_file(file_path)
        with pytest.raises(markup_utils.FileSizeError):
            markup_utils._get_file(self.directory, file_path)

    def test_get_doc_files(self):
        res = markup_utils.get_doc_files(self.directory)
        assert res is None

        for doc_file in [
            'GETTING_STARTED.md',
            'DEEP_DIVE.md',
            'WHOOPS.txt',
            'EXAMPLES.md',
        ]:
            self.fs.create_file(os.path.join(self.directory, doc_file))

        self.fs.create_dir(os.path.join(self.directory, 'sub_dir_to_ignore'))
        res = markup_utils.get_doc_files(self.directory)
        assert len(res) == 3
        names = [d.name for d in res]
        assert 'GETTING_STARTED.md' in names
        assert 'DEEP_DIVE.md' in names
        assert 'EXAMPLES.md' in names
        assert 'WHOOPS.txt' not in names

    def test_find_doc_files_no_dir(self):
        res = markup_utils._find_doc_files(directory='/does_not_exist')
        assert res == []


class TestHtmlRender(TestCase):
    def call_render(self, raw_text, mimetype):
        doc_file = DocFile(text=raw_text, mimetype=mimetype)
        return markup_utils._render_from_markdown(doc_file)

    def test_get_html(self):
        doc_file = DocFile(text=TEXT_SIMPLE, mimetype='text/markdown')
        html = markup_utils.get_html(doc_file)
        assert html == '<p>{}</p>'.format(TEXT_SIMPLE)

        doc_file = DocFile(text=TEXT_SIMPLE, mimetype='text/rst')
        html = markup_utils.get_html(doc_file)
        assert html is None

    def test_render_simple(self):
        html = self.call_render(TEXT_SIMPLE, 'text/markdown')
        assert html == '<p>{}</p>'.format(TEXT_SIMPLE)

    def test_render_bad_tag(self):
        html = self.call_render(TEXT_BAD_TAG, 'text/markdown')
        assert '<script>' not in html

    def test_render_bad_html_hidden_in_md(self):
        html = self.call_render(TEXT_BAD_HTML_IN_MD, 'text/markdown')
        assert 'javascript' not in html

    def test_render_formatting(self):
        html = self.call_render(TEXT_FORMATTING, 'text/markdown')
        assert '<h1>Role</h1>' in html
        assert '<a href="https://www.example.com">Tool</a>' in html
        assert '<code>$PATH</code>' in html
        assert '<h3>Installation</h3>' in html
        assert '<code>package_version: "1.2.0"' in html
        assert '<blockquote>\n<p>NOTE:' in html
        assert 'Tool \'feature\' is <em>beta</em>' in html
        assert '<ul>\n<li>Item1</li>' in html



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

from collections import namedtuple
import json
import os
import tempfile
from unittest import mock

import attr
import pytest

from galaxy_importer import collection
from galaxy_importer.collection import CollectionLoader
from galaxy_importer.config import ConfigFile
from galaxy_importer.constants import ContentType
from galaxy_importer import exceptions as exc
from galaxy_importer import schema
from galaxy_importer.utils import markup as markup_utils


CollectionFilename = \
    namedtuple("CollectionFilename", ["namespace", "name", "version"])

MANIFEST_JSON = """
{
 "collection_info": {
  "namespace": "my_namespace",
  "name": "my_collection",
  "version": "2.0.2",
  "authors": [
   "John Doe"
  ],
  "readme": "README.md",
  "tags": [
   "deployment",
   "fedora"
  ],
  "description": "A collection with various roles and plugins",
  "license": [
   "GPL-3.0-or-later",
   "MIT"
  ],
  "license_file": null,
  "dependencies": {
   "my_namespace.collection_nginx": ">=0.1.6",
   "network_user.collection_inspect": "2.0.0",
   "dave.deploy": "*"
  },
  "repository": null,
  "documentation": null,
  "homepage": null,
  "issues": null
 }
}
"""


@pytest.fixture
def tmp_collection_root():
    import shutil
    try:
        tmp_dir = tempfile.TemporaryDirectory().name
        sub_path = 'ansible_collections/placeholder_namespace/placeholder_name'
        collection_root = os.path.join(tmp_dir, sub_path)
        os.makedirs(collection_root)
        yield collection_root
    finally:
        # TODO: check if this applies to dir not just file
        shutil.rmtree(tmp_dir)


@mock.patch('galaxy_importer.collection.CollectionLoader._build_docs_blob')
def test_manifest_success(_build_docs_blob, tmp_collection_root):
    _build_docs_blob.return_value = {}
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(MANIFEST_JSON)

    with open(os.path.join(tmp_collection_root, 'README.md'), 'w'):
        pass

    filename = CollectionFilename('my_namespace', 'my_collection', '2.0.2')
    data = CollectionLoader(tmp_collection_root, filename).load()
    assert data.metadata.namespace == 'my_namespace'
    assert data.metadata.name == 'my_collection'
    assert data.metadata.version == '2.0.2'
    assert data.metadata.authors == ['John Doe']
    assert data.metadata.readme == 'README.md'
    assert data.metadata.tags == ['deployment', 'fedora']
    assert data.metadata.description == \
        'A collection with various roles and plugins'
    assert data.metadata.license_file is None
    assert data.metadata.dependencies == {
        'my_namespace.collection_nginx': '>=0.1.6',
        'network_user.collection_inspect': '2.0.0',
        'dave.deploy': '*'
    }
    assert data.metadata.repository is None
    assert data.metadata.homepage is None
    assert data.metadata.issues is None


@pytest.mark.parametrize(
    'manifest_text,new_text,error_subset',
    [
        ('my_namespace', '', "'namespace' is required"),
        ('my_namespace', '00my.name.space', "'namespace' has invalid format"),
        ('my_collection', '', "'name' is required"),
        ('my_collection', '_my_collection', "'name' has invalid format"),
        ('2.0.2', '', "'version' is required"),
        ('2.0.2', '2.2.0.0.2', "semantic version format"),
        ('"John Doe"', '', "'authors' is required"),
        ('[\n   "John Doe"\n  ]', '"John Doe"', "to be a list of strings"),
        ('README.md', '', "'readme' is required"),
        ('"fedora"', '["fedora"]', "to be a list of strings"),
        ('"deployment",', '"tag",'*30, "Expecting no more than 20 tags"),
        ('"A collection with various roles and plugins"', '[]', "be a string"),
        ('"MIT"', '{}', "to be a list of strings"),
        ('"MIT"', '"not-a-valid-license-id"', "list of valid SPDX license"),
        ('"*"', '555', "Expecting depencency version to be string"),
        ('"dave.deploy"', '"davedeploy"', "Invalid dependency format:"),
        ('"dave.deploy"', '"007.deploy"', "Invalid dependency format: '007'"),
        ('"dave.deploy"', '"my_namespace.my_collection"', "self dependency"),
        ('"*"', '"3.4.0.4"', "version spec range invalid"),
        ('"repository": null', '"repository": []', "must be a string"),
        ('"documentation": null', '"documentation": []', "must be a string"),
        ('"homepage": null', '"homepage": []', "must be a string"),
        ('"issues": null', '"issues": []', "must be a string"),
    ],
)
def test_manifest_fail(manifest_text, new_text, error_subset, tmp_collection_root):
    manifest_edited = MANIFEST_JSON.replace(manifest_text, new_text)
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(manifest_edited)

    with pytest.raises(exc.ManifestValidationError,
                       match=error_subset):
        CollectionLoader(
            tmp_collection_root, 'my_namespace-my_collection-2.0.2.tar.gz').load()


def test_build_contents_blob():
    collection_loader = CollectionLoader('/tmpdir', 'filename')
    collection_loader.content_objs = [
        schema.Content(name='my_module', content_type=ContentType.MODULE),
        schema.Content(name='my_role', content_type=ContentType.ROLE),
    ]
    res = collection_loader._build_contents_blob()
    assert [attr.asdict(item) for item in res] == [
        {'content_type': 'module', 'description': None, 'name': 'my_module'},
        {'content_type': 'role', 'description': None, 'name': 'my_role'}
    ]


@mock.patch('galaxy_importer.utils.markup.get_html')
@mock.patch('galaxy_importer.utils.markup.get_readme_doc_file')
def test_build_docs_blob_contents(get_readme_doc_file, get_html):
    get_readme_doc_file.return_value.name = 'README.md'
    get_html.return_value = '<p>A detailed guide</p>'
    collection_loader = CollectionLoader('/tmpdir', 'filename')
    collection_loader.content_objs = [
        schema.Content(name='my_module', content_type=ContentType.MODULE),
        schema.Content(name='my_role', content_type=ContentType.ROLE),
    ]
    res = collection_loader._build_docs_blob()
    assert attr.asdict(res) == {
        'collection_readme': {'name': 'README.md',
                              'html': '<p>A detailed guide</p>'},
        'documentation_files': [],
        'contents': [
            {
                'content_name': 'my_module',
                'content_type': 'module',
                'doc_strings': {},
                'readme_file': None,
                'readme_html': None,
            },
            {
                'content_name': 'my_role',
                'content_type': 'role',
                'doc_strings': {},
                'readme_file': None,
                'readme_html': None,
            },
        ],
    }


@mock.patch('galaxy_importer.utils.markup.get_html')
@mock.patch('galaxy_importer.utils.markup.get_readme_doc_file')
@mock.patch('galaxy_importer.utils.markup.get_doc_files')
def test_build_docs_blob_doc_files(get_doc_files, get_readme, get_html):
    get_readme.return_value.name = 'README.md'
    get_html.return_value = '<p>A detailed guide</p>'
    get_doc_files.return_value = [
        markup_utils.DocFile(name='INTRO.md', text='Intro text',
                             mimetype='text/markdown', hash=''),
        markup_utils.DocFile(name='INTRO2.md', text='Intro text',
                             mimetype='text/markdown', hash=''),
    ]
    collection_loader = CollectionLoader('/tmpdir', 'filename')
    collection_loader.content_objs = []
    res = collection_loader._build_docs_blob()
    assert attr.asdict(res) == {
        'collection_readme': {'name': 'README.md',
                              'html': '<p>A detailed guide</p>'},
        'documentation_files': [
            {
                'name': 'INTRO.md',
                'html': '<p>A detailed guide</p>',
            },
            {
                'name': 'INTRO2.md',
                'html': '<p>A detailed guide</p>',
            },
        ],
        'contents': [],
    }


@mock.patch('galaxy_importer.utils.markup.get_readme_doc_file')
def test_build_docs_blob_no_readme(get_readme_doc_file):
    get_readme_doc_file.return_value = None
    collection_loader = CollectionLoader('/tmpdir', 'filename')
    collection_loader.content_objs = []
    with pytest.raises(exc.ImporterError):
        collection_loader._build_docs_blob()


@mock.patch('galaxy_importer.collection.CollectionLoader._build_docs_blob')
def test_filename_empty_value(_build_docs_blob, tmp_collection_root):
    _build_docs_blob.return_value = {}
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(MANIFEST_JSON)
    with open(os.path.join(tmp_collection_root, 'README.md'), 'w'):
        pass

    filename = CollectionFilename(
        namespace='my_namespace',
        name='my_collection',
        version=None)
    data = CollectionLoader(tmp_collection_root, filename).load()
    assert data.metadata.namespace == 'my_namespace'
    assert data.metadata.name == 'my_collection'
    assert data.metadata.version == '2.0.2'


@mock.patch('galaxy_importer.collection.CollectionLoader._build_docs_blob')
def test_filename_none(_build_docs_blob, tmp_collection_root):
    _build_docs_blob.return_value = {}
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(MANIFEST_JSON)
    with open(os.path.join(tmp_collection_root, 'README.md'), 'w'):
        pass

    filename = None
    data = CollectionLoader(tmp_collection_root, filename).load()
    assert data.metadata.namespace == 'my_namespace'
    assert data.metadata.name == 'my_collection'
    assert data.metadata.version == '2.0.2'


def test_filename_not_match_metadata(tmp_collection_root):
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(MANIFEST_JSON)

    filename = CollectionFilename('diff_ns', 'my_collection', '2.0.2')
    with pytest.raises(exc.ManifestValidationError):
        CollectionLoader(tmp_collection_root, filename).load()


def test_license_file(tmp_collection_root):
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        manifest = json.loads(MANIFEST_JSON)
        manifest['collection_info']['license'] = []
        manifest['collection_info']['license_file'] = 'my_license.txt'
        fh.write(json.dumps(manifest))
    with open(os.path.join(tmp_collection_root, 'README.md'), 'w'):
        pass
    with open(os.path.join(tmp_collection_root, 'my_license.txt'), 'w'):
        pass
    data = CollectionLoader(tmp_collection_root, filename=None).load()
    assert data.metadata.license_file == 'my_license.txt'


def test_missing_readme(tmp_collection_root):
    with open(os.path.join(tmp_collection_root, 'MANIFEST.json'), 'w') as fh:
        fh.write(MANIFEST_JSON)
    with pytest.raises(
        exc.ManifestValidationError,
        match=r"Could not find file README.md"
    ):
        CollectionLoader(tmp_collection_root, filename=None).load()


def test_import_collection(mocker):
    mocker.patch.object(collection, '_import_collection')
    mocker.patch.object(ConfigFile, 'load')
    collection.import_collection(file=None, cfg=None)
    assert ConfigFile.load.called
    assert collection._import_collection.called



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import json
from unittest import mock

import pytest

from galaxy_importer import loaders


@pytest.fixture
def doc_string_loader():
    return loaders.DocStringLoader(
        path='/tmp_dir/tmp123/ansible_collections/my_namespace/my_collection',
        fq_collection_name='my_namespace.my_collection')


def test_init_loader(doc_string_loader):
    assert doc_string_loader.fq_collection_name == 'my_namespace.my_collection'


def test_get_plugins(doc_string_loader, tmpdir):
    tmpdir.join('__init__.py').write('')
    tmpdir.join('should_be_ignored.txt').write('')
    tmpdir.join('my_module.py').write('')
    plugins = doc_string_loader._get_plugins(str(tmpdir))
    assert plugins == ['my_namespace.my_collection.my_module']


def test_get_plugins_subdirs(doc_string_loader, tmpdir):
    tmpdir.mkdir('subdir1').mkdir('subdir2').join('nested_plugin.py').write('')
    plugins = doc_string_loader._get_plugins(str(tmpdir))
    assert plugins == ['my_namespace.my_collection.subdir1.subdir2.nested_plugin']


@mock.patch('galaxy_importer.loaders.Popen')
def test_run_ansible_doc(mocked_popen, doc_string_loader):
    mocked_popen.return_value.communicate.return_value = (
        '"expected output"', '')
    mocked_popen.return_value.returncode = 0
    res = doc_string_loader._run_ansible_doc(plugin_type='', plugins=[])
    assert res == 'expected output'


@mock.patch('galaxy_importer.loaders.Popen')
def test_run_ansible_doc_exception(mocked_popen, doc_string_loader):
    mocked_popen.return_value.communicate.return_value = (
        'output', 'error that causes exception')
    mocked_popen.return_value.returncode = 1
    res = doc_string_loader._run_ansible_doc(plugin_type='', plugins=[])
    assert not res


@mock.patch.object(loaders.DocStringLoader, '_run_ansible_doc')
def test_ansible_doc_no_output(mocked_run_ansible_doc, doc_string_loader):
    mocked_run_ansible_doc.return_value = ''
    assert doc_string_loader.load() == {}


def test_process_doc_strings_not_dict(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "return": {
                    "message": {
                        "description": ["The output message the sample module generates"]
                    },
                    "original_message": {
                        "description": ["The original name param that was passed in"],
                        "type": "str"
                    }
                }
            }
        }
    """

    data = json.loads(ansible_doc_output)
    res = doc_string_loader._process_doc_strings(data)
    assert res['my_sample_module']['return'] == [
        {
            'name': 'message',
            'description': ['The output message the sample module generates']
        },
        {
            'name': 'original_message',
            'description': ['The original name param that was passed in'],
            'type': 'str'
        }
    ]


def test_transform_doc_strings_return(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "return": {
                    "message": {
                        "description": ["The output message the sample module generates"]
                    },
                    "original_message": {
                        "description": ["The original name param that was passed in"],
                        "type": "str"
                    }
                }
            }
        }
    """

    data = json.loads(ansible_doc_output)
    data = list(data.values())[0]
    transformed_data = doc_string_loader._transform_doc_strings(data)
    assert transformed_data['return'] == [
        {
            'name': 'message',
            'description': ['The output message the sample module generates']
        },
        {
            'name': 'original_message',
            'description': ['The original name param that was passed in'],
            'type': 'str'
        }
    ]


def test_transform_doc_strings_options(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "doc": {
                    "description": ["Sample module for testing."],
                    "short_description": "Sample module for testing",
                    "version_added": "2.8",
                    "options": {
                        "exclude": {
                            "description": ["This is the message to send..."],
                            "required": "true"
                        },
                        "use_new": {
                            "description": ["Control is passed..."],
                            "version_added": "2.7",
                            "default": "auto"
                        }
                    }
                }
            }
        }
    """
    data = json.loads(ansible_doc_output)
    data = list(data.values())[0]
    transformed_data = doc_string_loader._transform_doc_strings(data)
    assert transformed_data['doc']['options'] == [
        {
            'name': 'exclude',
            'description': ['This is the message to send...'],
            'required': 'true',
        },
        {
            'name': 'use_new',
            'description': ['Control is passed...'],
            'version_added': '2.7',
            'default': 'auto'
        }
    ]


def test_transform_doc_strings_nested_contains(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "return": {
                    "resources": {
                        "contains": {
                            "acceleratorType": {
                                "description": ["The type of..."],
                                "returned": "success"
                            },
                            "networkEndpoints": {
                                "contains": {
                                    "ipAddress": {
                                        "description": ["The IP address."],
                                        "type": "str"
                                    },
                                    "port": {
                                        "description": ["The port"],
                                        "type": "int"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    """

    data = json.loads(ansible_doc_output)
    data = list(data.values())[0]
    transformed_data = doc_string_loader._transform_doc_strings(data)
    assert transformed_data['return'] == [
        {
            'name': 'resources',
            'contains': [
                {
                    'name': 'acceleratorType',
                    'description': ['The type of...'],
                    'returned': 'success',
                },
                {
                    'name': 'networkEndpoints',
                    'contains': [
                        {
                            'name': 'ipAddress',
                            'description': ['The IP address.'],
                            'type': 'str',
                        },
                        {
                            'name': 'port',
                            'description': ['The port'],
                            'type': 'int',
                        },
                    ]
                }
            ]
        },
    ]


def test_transform_doc_strings_nested_contains_dict_of_list(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "return": {
                    "output": {
                        "contains": {
                            "formatted_output": [
                                "Contains formatted response ..."
                            ]
                        }
                    }
                }
            }
        }
    """
    data = json.loads(ansible_doc_output)
    data = list(data.values())[0]
    transformed_data = doc_string_loader._transform_doc_strings(data)
    assert transformed_data['return'] == [
        {
            'name': 'output',
            'contains': [
            ]
        },
    ]


def test_transform_doc_strings_nested_suboptions(doc_string_loader):
    ansible_doc_output = """
        {
            "my_sample_module": {
                "doc": {
                    "description": ["Sample module for testing."],
                    "short_description": "Sample module for testing",
                    "version_added": "2.8",
                    "options": {
                        "exclude": {
                            "description": ["This is the message to send..."],
                            "required": "true"
                        },
                        "lan2_port_setting": {
                            "description": ["Control is passed..."],
                            "suboptions": {
                                "enabled": {
                                    "description": ["If set to True..."],
                                    "type": "bool"
                                },
                                "network_setting": {
                                    "description": ["If the enable field..."],
                                    "suboptions": {
                                        "address": {
                                            "description": ["The IPv4 Address of LAN2"]
                                        },
                                        "gateway": {
                                            "description": ["The default gateway of LAN2"]
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    """
    data = json.loads(ansible_doc_output)
    data = list(data.values())[0]
    transformed_data = doc_string_loader._transform_doc_strings(data)
    assert transformed_data['doc']['options'] == [
        {
            'name': 'exclude',
            'description': ['This is the message to send...'],
            'required': 'true',
        },
        {
            'name': 'lan2_port_setting',
            'description': ['Control is passed...'],
            'suboptions': [
                {
                    'name': 'enabled',
                    'description': ['If set to True...'],
                    'type': 'bool',
                },
                {
                    'name': 'network_setting',
                    'description': ['If the enable field...'],
                    'suboptions': [
                        {
                            'name': 'address',
                            'description': ['The IPv4 Address of LAN2'],
                        },
                        {
                            'name': 'gateway',
                            'description': ['The default gateway of LAN2'],
                        }
                    ]
                }
            ]
        }
    ]


@mock.patch.object(loaders.DocStringLoader, '_run_ansible_doc')
def test_load(mocked_run_ansible_doc, doc_string_loader, tmpdir):
    ansible_doc_output = """
        {
            "my_module": {
                "return": {
                    "message": {
                        "description": ["The output message the sample module generates"]
                    },
                    "original_message": {
                        "description": ["The original name param that was passed in"],
                        "type": "str"
                    }
                }
            }
        }
    """
    mocked_run_ansible_doc.return_value = json.loads(ansible_doc_output)

    doc_string_loader.path = str(tmpdir)
    tmpdir.mkdir('plugins').mkdir('modules').join('my_module.py').write('')

    res = doc_string_loader.load()
    assert res == {
        'module': {
            'my_module': {
                'return': [
                    {
                        'description': ['The output message the sample module generates'],
                        'name': 'message'
                    },
                    {
                        'description': ['The original name param that was passed in'],
                        'name': 'original_message', 'type': 'str'}
                ]
            }
        }
    }


@mock.patch('galaxy_importer.loaders.Popen')
def test_load_ansible_doc_error(mocked_popen, doc_string_loader, tmpdir):
    mocked_popen.return_value.communicate.return_value = (
        'output', 'error that causes exception')
    mocked_popen.return_value.returncode = 1

    doc_string_loader.path = str(tmpdir)
    tmpdir.mkdir('plugins').mkdir('inventory').join('my_plugin.py').write('')

    res = doc_string_loader.load()
    assert res == {'inventory': {}}



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import pytest

from galaxy_importer.schema import CollectionInfo


@pytest.fixture
def collection_info():
    metadata = {
        'namespace': 'acme',
        'name': 'jenkins',
        'version': '3.5.0',
        'license': ['MIT'],
        'readme': 'README.md',
        'authors': ['Bob Smith <b.smith@acme.com>'],
        'tags': ['testcases']
    }
    return metadata


def test_collection_info(collection_info):
    res = CollectionInfo(**collection_info)
    assert type(res) == CollectionInfo
    assert res.namespace == 'acme'
    assert res.name == 'jenkins'
    assert res.version == '3.5.0'
    assert res.license == ['MIT']
    assert res.readme == 'README.md'
    assert res.authors == ['Bob Smith <b.smith@acme.com>']
    assert res.tags == ['testcases']


def test_readme_req(collection_info):
    collection_info['readme'] = ''
    with pytest.raises(ValueError, match=r"'readme' is required"):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'valid_name',
    [
        'my_namespace',
        'my_name',
        'roles75',
        'nginx',
        'nginx_',
        'deploy4py_script',
    ]
)
def test_valid_names(collection_info, valid_name):
    collection_info['name'] = valid_name
    res = CollectionInfo(**collection_info)
    assert res.name == valid_name


@pytest.mark.parametrize(
    'invalid_name',
    [
        '_leading_underscore',
        'has-some-dashes',
        '5tarts_with_number',
        '030',
        '0x4e3',
        'has space',
        'hasUpperCase',
        'double__under',
        'invalid#char',
        'inv@lid/char',
        'no.dots',
    ]
)
def test_invalid_names(collection_info, invalid_name):
    collection_info['name'] = invalid_name
    with pytest.raises(ValueError, match=r"'name' has invalid format"):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'valid_tags',
    [
        ['good_tag', 'goodtag'],
        ['deployment'],
        ['fedora'],
        ['fedora29'],
        ['fedora_29'],
        ['alloneword'],
        ['deployment', 'fedora', 'a_007', 'alloneword']
    ]
)
def test_valid_tags(collection_info, valid_tags):
    collection_info['tags'] = valid_tags
    res = CollectionInfo(**collection_info)
    assert res.tags == valid_tags


@pytest.mark.parametrize(
    'invalid_tags',
    [
        ['007'],
        ['4ubuntu'],
        ['0x4e3'],
        ['bad-tag'],
        ['Badtag'],
        ['badTAG'],
        ['bad-tag', 'goodtag'],
        ['bad tag'],
        ['bad.tag'],
        ['bad.tag', 'bad tag'],
        ['_deploy'],
        ['inv@lid/char'],
    ]
)
def test_invalid_tags(collection_info, invalid_tags):
    collection_info['tags'] = invalid_tags
    with pytest.raises(ValueError, match="'tag' has invalid format"):
        CollectionInfo(**collection_info)


def test_max_tags(collection_info):
    collection_info['tags'] = [str(f'tag_{i}') for i in range(90, 110)]
    res = CollectionInfo(**collection_info)
    assert [str(f'tag_{i}') for i in range(90, 110)] == res.tags

    collection_info['tags'] = [str(f'tag_{i}') for i in range(90, 111)]
    with pytest.raises(ValueError, match=r'Expecting no more than '):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'valid_semver',
    [
        '1.2.3',
        '1.0.0-beta',
        '2.0.0-rc.2',
        '4.6.21',
        '0.1.1-alpha+build.2012-05-15',
    ]
)
def test_valid_semver(collection_info, valid_semver):
    collection_info['version'] = valid_semver
    res = CollectionInfo(**collection_info)
    assert res.version == valid_semver


@pytest.mark.parametrize(
    'invalid_semver',
    [
        '2',
        '1.2.3a',
        '2.0.02',
        '1.2.beta',
        '3,4',
        '3.4.0.4',
        'latest',
        'v0',
    ]
)
def test_invalid_semver(collection_info, invalid_semver):
    collection_info['version'] = invalid_semver
    with pytest.raises(ValueError,
                       match=r"'version' to be in semantic version"):
        CollectionInfo(**collection_info)


def test_license(collection_info):
    collection_info['license'] = ['MIT', 'GPL-2.0-only', 'Apache-2.0']
    res = CollectionInfo(**collection_info)
    assert res.license == ['MIT', 'GPL-2.0-only', 'Apache-2.0']
    assert res.license_file is None

    collection_info['license'] = ['MIT', 'not-a-real-license', 'n00p']
    with pytest.raises(ValueError, match=r"Expecting 'license' to be a list of valid SPDX"):
        CollectionInfo(**collection_info)

    collection_info['license'] = 'MIT'
    with pytest.raises(ValueError, match=r"'license' to be a list of strings"):
        CollectionInfo(**collection_info)


def test_license_file(collection_info):
    collection_info['license'] = []
    collection_info['license_file'] = 'my_very_own_license.txt'
    res = CollectionInfo(**collection_info)
    assert len(res.license) == 0
    assert res.license_file == 'my_very_own_license.txt'

    collection_info['license_file'] = ['my_very_own_license.txt']
    with pytest.raises(ValueError, match=r"'license_file' must be a string"):
        CollectionInfo(**collection_info)


def test_empty_lic_and_lic_file(collection_info):
    collection_info['license'] = []
    with pytest.raises(ValueError, match=r"'license' or 'license_file' are required"):
        CollectionInfo(**collection_info)


def test_both_lic_and_lic_file(collection_info):
    collection_info['license_file'] = 'my_very_own_license.txt'
    with pytest.raises(
        ValueError,
        match=r"'license' and 'license_file' keys are mutually exclusive"
    ):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'valid_license_list',
    [
        ['MIT'],
        ['Apache-2.0'],
        ['BSD-3-Clause'],
        ['CC-BY-4.0', 'MIT']
    ]
)
def test_valid_license(collection_info, valid_license_list):
    collection_info['license'] = valid_license_list
    res = CollectionInfo(**collection_info)
    assert res.license == valid_license_list


@pytest.mark.parametrize(
    'invalid_license_list',
    [
        ['BSD'],
        ['Apache'],
        ['MIT AND Apache-2.0'],
        ['something_else'],
        ['CC-BY-4.0', 'MIT', 'bad_license_id']
    ]
)
def test_invalid_license(collection_info, invalid_license_list):
    collection_info['license'] = invalid_license_list
    with pytest.raises(ValueError, match=r'invalid license identifiers'):
        CollectionInfo(**collection_info)


def test_invalid_dep_type(collection_info):
    collection_info['dependencies'] = 'joe.role1: 3'
    with pytest.raises(TypeError,
                       match=r"'dependencies' must be <class 'dict'>"):
        CollectionInfo(**collection_info)


def test_invalid_dep_name(collection_info):
    collection_info['dependencies'] = {3.3: '1.0.0'}
    with pytest.raises(ValueError, match=r'Expecting depencency to be string'):
        CollectionInfo(**collection_info)


def test_invalid_dep_version(collection_info):
    collection_info['dependencies'] = {'joe.role1': 3}
    with pytest.raises(ValueError,
                       match=r'Expecting depencency version to be string'):
        CollectionInfo(**collection_info)


def test_non_null_str_fields(collection_info):
    collection_info['description'] = None
    res = CollectionInfo(**collection_info)
    assert res.description is None

    collection_info['description'] = 'description of the collection'
    res = CollectionInfo(**collection_info)
    assert res.description == 'description of the collection'

    collection_info['description'] = ['should be a string not list']
    with pytest.raises(ValueError, match=r"'description' must be a string"):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'dependent_collection',
    [
        'no_dot_in_collection',
        'too.many.dots',
        '.too.many.dots',
        'too.many.dots.',
        'empty_name.',
        '.empty_namespace',
        'a_user._leading_underscore',
        '5tarts_with_number.gunicorn',
        'mynamespace.UPPERCASE_COLLECTION',
        'MyNamespace.mixedcase_collection',
        'mynamespace.my-dashed-collection',
        'my-namespace.mydashedcollection',
        'mynamespace.my spaced collection',
        'my namespace.myspacedcollection',
    ]
)
def test_invalid_dep_format(collection_info, dependent_collection):
    collection_info['dependencies'] = {dependent_collection: '1.0.0'}
    with pytest.raises(ValueError, match=r'Invalid dependency format'):
        CollectionInfo(**collection_info)


@pytest.mark.parametrize(
    'depencency',
    [
        {'alice.apache': 'bad_version'},
        {'alice.apache': ''},
        {'alice.apache': '>=1.0.0, <=2.0.0'},
        {'alice.apache': '>1 <2'},
    ]
)
def test_invalid_version_spec(collection_info, depencency):
    collection_info['dependencies'] = depencency
    with pytest.raises(ValueError, match='version spec range invalid'):
        CollectionInfo(**collection_info)


def test_self_dependency(collection_info):
    namespace = collection_info['namespace']
    name = collection_info['name']
    collection_info['dependencies'] = {
        '{}.{}'.format(namespace, name): '1.0.0'
    }
    with pytest.raises(ValueError, match=r'Cannot have self dependency'):
        CollectionInfo(**collection_info)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import json
import os
import re
import shutil
import tempfile
from types import SimpleNamespace
from unittest import mock

import attr
import pytest

from galaxy_importer import constants
from galaxy_importer import exceptions as exc
from galaxy_importer import loaders
from galaxy_importer import schema


ANSIBLE_DOC_OUTPUT = """
    {"module": {
        "my_ns.my_collection.my_module": {
            "doc": {
                "description": ["Sample module for testing."],
                "short_description": "Sample module for testing",
                "version_added": "2.8",
                "options": {
                    "exclude": {
                        "description": ["This is the message to send..."],
                        "required": "true"
                    },
                    "use_new": {
                        "description": ["Control is passed..."],
                        "version_added": "2.7",
                        "default": "auto"
                    }
                }
            },
            "examples": null,
            "metadata": null,
            "return": {
                "message": {
                    "description": "The output message the sample module generates"
                },
                "original_message": {
                    "description": "The original name param that was passed in",
                    "type": "str"
                }
            }
        },
        "my_ns.my_collection.subdir1.subdir2.my_module_2": {
            "doc": {
                "short_description": "Module with subdirs",
                "version_added": "2.9"
            }
        }
    }}
"""


@pytest.fixture
def loader_module():
    return loaders.PluginLoader(
        content_type=constants.ContentType.MODULE,
        rel_path='plugins/modules/my_module.py',
        root='/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection',
        cfg=SimpleNamespace(run_flake8=True),
        doc_strings=json.loads(ANSIBLE_DOC_OUTPUT))


@pytest.fixture
def loader_doc_fragment():
    return loaders.PluginLoader(
        content_type=constants.ContentType.DOC_FRAGMENTS_PLUGIN,
        rel_path='plugins/doc_fragments/my_doc_fragment.py',
        root='/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection',
        cfg=SimpleNamespace(run_flake8=False),
        doc_strings=json.loads(ANSIBLE_DOC_OUTPUT))


@pytest.fixture
def loader_role():
    return loaders.RoleLoader(
        content_type=constants.ContentType.ROLE,
        rel_path='roles/my_sample_role',
        root='/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection')


@pytest.fixture
def loader_module_subdirs():
    return loaders.PluginLoader(
        content_type=constants.ContentType.MODULE,
        rel_path='plugins/modules/subdir1/subdir2/my_module_2.py',
        root='/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection',
        cfg=SimpleNamespace(run_flake8=False),
        doc_strings=json.loads(ANSIBLE_DOC_OUTPUT))


@pytest.fixture
def loader_role_subdirs():
    return loaders.RoleLoader(
        content_type=constants.ContentType.ROLE,
        rel_path='roles/subdir1/subdir2/my_sample_role',
        root='/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection')


def test_get_loader_cls():
    res = loaders.get_loader_cls(constants.ContentType.MODULE)
    assert issubclass(res, loaders.PluginLoader)
    assert not issubclass(res, loaders.RoleLoader)

    res = loaders.get_loader_cls(constants.ContentType.ROLE)
    assert issubclass(res, loaders.RoleLoader)
    assert not issubclass(res, loaders.PluginLoader)


def test_init_plugin_loader(loader_module):
    assert loader_module.name == 'my_module'
    assert loader_module.path_name == 'my_module'


def test_init_role_loader(loader_role):
    assert loader_role.name == 'my_sample_role'
    assert loader_role.path_name == 'my_sample_role'


def test_init_plugin_loader_subdirs(loader_module_subdirs):
    assert loader_module_subdirs.name == 'my_module_2'
    assert loader_module_subdirs.path_name == 'subdir1.subdir2.my_module_2'


def test_init_role_loader_subdirs(loader_role_subdirs):
    assert loader_role_subdirs.name == 'my_sample_role'
    assert loader_role_subdirs.path_name == 'subdir1.subdir2.my_sample_role'


def test_bad_plugin_name():
    with pytest.raises(exc.ContentNameError):
        loaders.PluginLoader(
            content_type=constants.ContentType.MODULE,
            rel_path='plugins/modules/bad-name-dashes.py',
            root='')


def test_bad_role_name():
    with pytest.raises(exc.ContentNameError):
        loaders.RoleLoader(
            content_type=constants.ContentType.ROLE,
            rel_path='roles/bad-name-dashes',
            root='')


def test_get_fq_collection_name(loader_module):
    root = '/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection'
    assert loader_module._get_fq_collection_name(root) == 'my_ns.my_collection'


def test_get_fq_name(loader_module):
    root = '/tmp_placeholder/tmp_placeholder/ansible_collections/my_ns/my_collection'
    res = loader_module._get_fq_name(root, 'subdir.my_module')
    assert res == 'my_ns.my_collection.subdir.my_module'


@mock.patch('galaxy_importer.loaders.Popen')
def test_plugin_loader_annotated_type(mocked_popen, loader_module):
    mocked_popen.return_value.stdout = ['my flake8 warning']
    assert loader_module.name == 'my_module'
    res = loader_module.load()
    assert isinstance(res, schema.Content)
    assert isinstance(res.content_type, attr.fields(schema.Content).content_type.type)


@mock.patch('galaxy_importer.loaders.Popen')
def test_load(mocked_popen, loader_module):
    mocked_popen.return_value.stdout = ''
    assert loader_module.name == 'my_module'
    res = loader_module.load()
    assert isinstance(res, schema.Content)
    assert res.name == 'my_module'
    assert res.content_type == constants.ContentType.MODULE
    assert res.description == 'Sample module for testing'
    assert res.doc_strings['doc']['version_added'] == '2.8'
    assert res.readme_file is None
    assert res.readme_html is None


def test_load_subdirs(loader_module_subdirs):
    assert loader_module_subdirs.name == 'my_module_2'
    res = loader_module_subdirs.load()
    assert isinstance(res, schema.Content)
    assert res.name == 'subdir1.subdir2.my_module_2'
    assert res.content_type == constants.ContentType.MODULE
    assert res.description == 'Module with subdirs'
    assert res.doc_strings['doc']['version_added'] == '2.9'
    assert res.readme_file is None
    assert res.readme_html is None


def test_load_doc_fragment_no_doc_strings(loader_doc_fragment):
    assert loader_doc_fragment.name == 'my_doc_fragment'
    res = loader_doc_fragment.load()
    assert isinstance(res, schema.Content)
    assert res.name == 'my_doc_fragment'
    assert res.content_type == constants.ContentType.DOC_FRAGMENTS_PLUGIN
    assert res.description is None
    assert res.doc_strings is None
    assert res.readme_file is None
    assert res.readme_html is None


@mock.patch('galaxy_importer.loaders.Popen')
def test_flake8_output(mocked_popen, loader_module):
    mocked_popen.return_value.stdout = ['my flake8 warning']
    res = list(loader_module._run_flake8('.'))
    assert res[0] == 'my flake8 warning'


ANSIBLELINT_TASK_OK = """---
- name: Add mongodb repo apt_key
  become: true
  apt_key: keyserver=hkp
  until: result.rc == 0
"""

ANSIBLELINT_TASK_SUDO_WARN = """---
- name: edit vimrc
  sudo: true
  lineinfile:
    path: /etc/vimrc
    line: '# added via ansible'
"""

ROLE_METADATA = """---
galaxy_info:
  description: Test description inside metadata
  license: MIT
"""

ROLE_METADATA_MISSING_DESC = """---
galaxy_info:
  license: MIT
"""


@pytest.fixture
def temp_root():
    try:
        tmp = tempfile.mkdtemp()
        yield tmp
    finally:
        shutil.rmtree(tmp)


def test_ansiblelint_file(loader_role):
    loader_role.root = None
    with tempfile.NamedTemporaryFile('w') as fp:
        fp.write(ANSIBLELINT_TASK_SUDO_WARN)
        fp.flush()
        res = list(loader_role._lint_role(fp.name))
    assert 'deprecated sudo' in ' '.join(res).lower()


def test_ansiblelint_role(temp_root, loader_role):
    task_dir = os.path.join(temp_root, 'tasks')
    loader_role.root = None
    os.makedirs(task_dir)
    with open(os.path.join(task_dir, 'main.yml'), 'w') as fp:
        fp.write(ANSIBLELINT_TASK_SUDO_WARN)
        fp.flush()
        res = list(loader_role._lint_role(temp_root))
    assert 'deprecated sudo' in ' '.join(res).lower()
    lint_output_path = re.compile(r'^tmp.+\/tasks\/main\.yml.+$')
    assert re.match(lint_output_path, res[0])


def test_ansiblelint_role_no_warn(temp_root, loader_role):
    loader_role.root = None
    task_dir = os.path.join(temp_root, 'tasks')
    os.makedirs(task_dir)
    with open(os.path.join(task_dir, 'main.yml'), 'w') as fp:
        fp.write(ANSIBLELINT_TASK_OK)
        fp.flush()
        res = list(loader_role._lint_role(temp_root))
    assert res == []


@mock.patch('galaxy_importer.loaders.Popen')
def test_ansible_lint_exception(mocked_popen, loader_role):
    mocked_popen.return_value.stdout = ''
    mocked_popen.return_value.wait.return_value = 1
    res = list(loader_role._lint_role('.'))
    assert 'Exception running ansible-lint' in res[0]


def test_find_metadata_file_path(temp_root, loader_role):
    root, rel_path = os.path.split(temp_root)

    res = loader_role._find_metadata_file_path(root, rel_path)
    assert res is None

    meta_dir = os.path.join(temp_root, 'meta')
    os.mkdir(meta_dir)
    with open(os.path.join(meta_dir, 'main.yml'), 'w'):
        pass
    res = loader_role._find_metadata_file_path(root, rel_path)
    assert res == os.path.join(temp_root, 'meta', 'main.yml')


def test_get_role_metadata_desc(temp_root, loader_role):
    loader_role.rel_path = temp_root
    meta_dir = os.path.join(temp_root, 'meta')
    os.mkdir(meta_dir)

    res = loader_role._get_metadata_description()
    assert res is None

    with open(os.path.join(meta_dir, 'main.yml'), 'w') as fp:
        fp.write(ROLE_METADATA)
    res = loader_role._get_metadata_description()
    assert res == 'Test description inside metadata'

    with open(os.path.join(meta_dir, 'main.yml'), 'w') as fp:
        fp.write(ROLE_METADATA_MISSING_DESC)
    res = loader_role._get_metadata_description()
    assert res is None


def test_get_role_readme(temp_root, loader_role):
    loader_role.root = temp_root
    loader_role.rel_path = ''
    with open(os.path.join(temp_root, 'README.md'), 'w') as fp:
        fp.write('This is the role readme text')
    readme = loader_role._get_readme()
    assert readme.name == 'README.md'
    assert readme.text == 'This is the role readme text'


def test_get_role_readme_fail(temp_root, loader_role):
    loader_role.root = temp_root
    loader_role.rel_path = ''
    with pytest.raises(exc.ContentLoadError):
        loader_role._get_readme()


@mock.patch.object(loaders.RoleLoader, '_lint_role')
def test_load_role(mocked_lint_role, temp_root, loader_role):
    mocked_lint_role.return_value = 'ANSIBLE_LINT_OUTPUT'
    loader_role.root = ''
    loader_role.rel_path = temp_root

    with open(os.path.join(temp_root, 'README.md'), 'w') as fp:
        fp.write('This is the role readme text')
    with open(os.path.join(temp_root, 'meta.yml'), 'w') as fp:
        fp.write(ROLE_METADATA)

    res = loader_role.load()
    assert isinstance(res, schema.Content)
    assert res.name == 'my_sample_role'
    assert res.content_type == constants.ContentType.ROLE
    assert res.description == 'Test description inside metadata'



# (c) 2012-2020, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging
import os
import subprocess
from types import SimpleNamespace

import pytest

from galaxy_importer import config
from galaxy_importer.ansible_test import runners


@pytest.fixture
def temp_config_file():
    try:
        dir = os.path.dirname(os.path.dirname(__file__))
        config_file = os.path.join(dir, 'galaxy_importer', 'galaxy-importer.cfg')
        yield config_file
    finally:
        os.remove(config_file)


def test_get_runner_no_config_file():
    config_data = config.ConfigFile.load()
    cfg = config.Config(config_data=config_data)
    assert runners.get_runner(cfg) is None


def test_get_runner_ansible_test_local(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert runners.get_runner(cfg) == runners.LocalAnsibleTestRunner


def test_get_runner_pulp(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True\n'
                'INFRA_PULP = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert runners.get_runner(cfg) == runners.LocalImageTestRunner


def test_get_runner_pulp_and_osd(temp_config_file):
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True\n'
                'INFRA_PULP = True\nINFRA_OSD = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)
        assert runners.get_runner(cfg) == runners.OpenshiftJobTestRunner


def test_ansible_test_runner_run(mocker, temp_config_file):
    mocker.patch.object(runners, 'LocalAnsibleTestRunner')
    mocker.patch.object(runners, 'OpenshiftJobTestRunner')
    with open(temp_config_file, 'w') as f:
        f.write('[galaxy-importer]\nRUN_ANSIBLE_TEST = True\n'
                'INFRA_PULP = True\nINFRA_OSD = True')
        f.flush()
        config_data = config.ConfigFile.load()
        cfg = config.Config(config_data=config_data)

        ansible_test_runner = runners.get_runner(cfg)
        ansible_test_runner().run()
        assert not runners.LocalAnsibleTestRunner.called
        assert runners.OpenshiftJobTestRunner.called


def test_local_run(mocker, caplog):
    mocker.patch.object(subprocess, 'Popen')
    subprocess.Popen.return_value.stdout = ['stdout_result']
    subprocess.Popen.return_value.wait.return_value = 0
    caplog.set_level(logging.INFO)

    metadata = SimpleNamespace(
        namespace='test_ns', name='test_name', version='test_version')
    runner = runners.local_ansible_test.LocalAnsibleTestRunner(metadata=metadata)
    runner.run()

    assert len(caplog.records) == 4
    assert subprocess.Popen.called
    assert 'stdout_result' in str(caplog.records[0])
    assert 'Running ansible-test sanity on test_ns-test_name-test_version' in \
        str(caplog.records[1])
    assert 'stdout_result' in str(caplog.records[3])


def test_local_run_rc_error(mocker, caplog):
    mocker.patch.object(subprocess, 'Popen')
    subprocess.Popen.return_value.stdout = ['stdout_result']
    subprocess.Popen.return_value.wait.return_value = 1
    caplog.set_level(logging.INFO)

    metadata = SimpleNamespace(
        namespace='test_ns', name='test_name', version='test_version')
    runner = runners.local_ansible_test.LocalAnsibleTestRunner(metadata=metadata)
    runner.run()

    assert subprocess.Popen.called
    assert len(caplog.records) == 5
    assert caplog.records[4].levelname == 'ERROR'
    assert 'An exception occurred in ansible-test sanity' in str(caplog.records[4])
